@@ -1,11 +1,12 @@
 ;;; mct.el --- Minibuffer Confines Transcended -*- lexical-binding: t -*-
 
-;; Copyright (C) 2021-2023  Free Software Foundation, Inc.
+;; Copyright (C) 2021-2025  Free Software Foundation, Inc.
 
 ;; Author: Protesilaos Stavrou <info@protesilaos.com>
-;; URL: https://git.sr.ht/~protesilaos/mct
-;; Version: 1.0.0
-;; Package-Requires: ((emacs "29.1"))
+;; Maintainer: Protesilaos Stavrou <info@protesilaos.com>
+;; URL: https://github.com/protesilaos/mct
+;; Version: 1.1.0
+;; Package-Requires: ((emacs "29.4"))
 
 ;; This file is NOT part of GNU Emacs.
 
@@ -24,23 +25,28 @@
 
 ;;; Commentary:
 ;;
-;; Enhancements for the default minibuffer completion UI of Emacs.  In
-;; essence, MCT is (i) a very thin layer of interactivity on top of the
-;; out-of-the-box completion experience, and (ii) glue code that combines
-;; built-in functionalities to make the default completion framework work
-;; like that of more featureful third-party options.
+;; Opinionated changes and enhancements to the default minibuffer
+;; completion UI of Emacs:
 ;;
-;; + Package name (GNU ELPA): `mct`
-;; + Official manual: <https://protesilaos.com/emacs/mct>
-;; + Change log: <https://protesilaos.com/emacs/mct-changelog>
-;; + Git repo on SourceHut: <https://git.sr.ht/~protesilaos/mct>
-;;   - Mirrors:
-;;     + GitHub: <https://github.com/protesilaos/mct>
-;;     + GitLab: <https://gitlab.com/protesilaos/mct>
-;; + Mailing list: <https://lists.sr.ht/~protesilaos/general-issues>
-;; + Video demo: <https://protesilaos.com/codelog/2021-10-22-emacs-mct-demo/>
-;; + Backronym: Minibuffer Confines Transcended; Minibuffer and
-;;   Completions in Tandem.
+;; - Live completions to update the results as you type.
+;;
+;; - Passlist and blocklist of commands or completion categories to
+;;   control whether the *Completions* buffer shows up eagerly.
+;;
+;; - Other relevant options to control when the *Completions* are
+;;   displayed.
+;;
+;; - Per command or completion category sorting methods.
+;;
+;; - A cleaner *Completions* buffer, optionally without a mode line.
+;;
+;; - Commands to cycle from the minibuffer to the *Completions* and
+;;   vice versa.
+;;
+;; In essence, MCT is (i) a layer of interactivity on top of the
+;; out-of-the-box completion experience, and (ii) glue code that
+;; combines built-in functionalities to make the default completion
+;; framework work like that of more featureful third-party options.
 
 ;;; Code:
 
@@ -50,7 +56,10 @@
   "Minibuffer Confines Transcended.
 A layer of interactivity that integrates the standard minibuffer
 and the Completions."
-  :group 'minibuffer)
+  :group 'minibuffer
+  :link '(info-link "(mct) Top")
+  :link '(url-link :tag "Homepage" "https://protesilaos.com/emacs/mct"))
+
 
 (make-obsolete 'mct-completion-windows-regexp 'mct--completions-window-name "0.5.0")
 
@@ -142,6 +151,20 @@ number of candidates that are being computed."
   :type 'natnum
   :group 'mct)
 
+(defcustom mct-completing-read-multiple-indicator t
+  "When non-nil show an indicator for `completing-read-multiple' prompts.
+If nil, do not show anything.  Those prompts will look like the generic ones.
+
+The indicator informs the user this is a `completing-read-multiple'
+prompt and also shows the `crm-separator', which is usually a comma.
+
+The `mct-completing-read-multiple-indicator' has no effect on versions
+of Emacs >= 31, as those have the user option `crm-prompt' which covers
+the same use-case."
+  :type 'boolean
+  :package-version '(mct . "1.1.0")
+  :group 'mct)
+
 (defcustom mct-live-update-delay 0.3
   "Delay in seconds before updating the Completions buffer.
 Set this to 0 to disable the delay.
@@ -195,51 +218,48 @@ Read the manual for known completion categories."
   :type '(repeat symbol)
   :group 'mct)
 
+(defcustom mct-sort-by-command-or-category
+  '((file . mct-sort-by-directory-then-by-file)
+    ((magit-checkout vc-retrieve-tag) . mct-sort-by-alpha-then-by-length)
+    ((kill-ring imenu consult-location Info-goto-node Info-index Info-menu) . nil)
+    (t . mct-sort-by-history))
+  "Sort completion candidates based on the command or completion category.
+This is an alist where each element is of the form (SYMBOLS . SORT-FUNCTION).
+
+SYMBOLS is either a symbol or a list of symbols.  SYMBOLS can refer to
+the symbol of a function or completion category.  It can also be t,
+which refers to the fallback value.
+
+SORT-FUNCTION is a function that takes a list of strings and returns a
+list of strings, sorting them accordingly.  Examples of a SORT-FUNCTION
+are:
+
+- `mct-sort-by-alpha'
+- `mct-sort-by-alpha-then-by-length'
+- `mct-sort-by-history'
+- `mct-sort-by-directory-then-by-file'
+
+To not perform any sorting on the completion candidates that match
+SYMBOLS set SORT-FUNCTION to nil.
+
+If there are conflicting configurations between a SYMBOLS function and
+completion category, then the function takes precedence (for example
+`find-file' is set to sort directories first whereas the `file'
+completion category is set to sort by history)."
+  :type '(alist
+          :key-type (choice symbol (repeat symbol))
+          :value-type (choice
+                       (const :tag "Sort A-Z" mct-sort-by-alpha)
+                       (const :tag "Sort A-Z then from short to long" mct-sort-by-alpha-then-by-length)
+                       (const :tag "Sort by minibuffer history" mct-sort-by-history)
+                       (const :tag "Sort by directory then by file" mct-sort-by-directory-then-by-file)
+                       (function :tag "Custom sort function accepting COMPLETIONS argument")))
+  :package-version '(mct . "1.1.0")
+  :group 'mct)
+
 (make-obsolete-variable 'mct-display-buffer-action nil "1.0.0")
 (make-obsolete-variable 'mct-completions-format nil "1.0.0")
-
-(defcustom mct-persist-dynamic-completion t
-  "When non-nil, keep dynamic completion live.
-
-Without any intervention from MCT, the default Emacs behavior for
-commands such as `find-file' or for a `file' completion category
-is to hide the `*Completions*' buffer after updating the list of
-candidates in a non-exiting fashion (e.g. select a directory and
-expect to continue typing the path).  This, however, runs
-contrary to the interaction model of MCT when it performs live
-completions, because the user expects the Completions buffer to
-remain visible while typing out the path to the file.
-
-When this user option is non-nil (the default) it makes all
-non-exiting commands keep the `*Completions*' visible when
-updating the list of candidates.
-
-This applies to prompts in the `file' completion category
-whenever the user selects a candidate with
-`mct-choose-completion-no-exit', `mct-edit-completion',
-`minibuffer-complete', `minibuffer-force-complete' (i.e. any
-command that does not exit the minibuffer).
-
-The two exceptions are (i) when the current completion session
-runs a command or category that is blocked by the
-`mct-completion-blocklist' or (ii) the user option
-`mct-live-completion' is nil.
-
-The underlying rationale:
-
-Most completion commands present a flat list of candidates to
-choose from.  Picking a candidate concludes the session.  Some
-prompts, however, can recalculate the list of completions based
-on the selected candidate.  A case in point is `find-file' (or
-any command with the `file' completion category) which
-dynamically adjusts the completions to show only the elements
-which extend the given file system path.  We call such cases
-\"dynamic completion\".  Due to their particular nature, these
-need to be handled explicitly.  The present user option is
-provided primarily to raise awareness about this state of
-affairs."
-  :type 'boolean
-  :group 'mct)
+(make-obsolete-variable 'mct-persist-dynamic-completion 'mct-completion-passlist "1.1.0")
 
 ;;;; Completion metadata
 
@@ -249,7 +269,7 @@ affairs."
 
 (defun mct--completion-category ()
   "Return completion category."
-  (when-let ((window (active-minibuffer-window)))
+  (when-let* ((window (active-minibuffer-window)))
     (with-current-buffer (window-buffer window)
       (completion-metadata-get
        (completion-metadata (buffer-substring-no-properties
@@ -272,43 +292,77 @@ affairs."
   "Return non-nil if symbol is in the `mct-completion-blocklist'."
   (mct--symbol-in-list mct-completion-blocklist))
 
-;; Normally we would also include `imenu', but it has its own defcustom
-;; for popping up the Completions eagerly...  Let's not interfere with
-;; that.
-;;
-;; See bug#52389: <https://debbugs.gnu.org/cgi/bugreport.cgi?bug=52389>.
-(defvar mct--dynamic-completion-categories '(file)
-  "Completion categories that perform dynamic completion.")
-
 ;;;; Sorting functions for `completions-sort' (Emacs 29)
 
-(defun mct-sort-by-alpha-length (elements)
-  "Sort ELEMENTS first alphabetically, then by length.
+(defvar mct-sort-alpha-function #'string-version-lessp
+  "Function to perform alphabetic sorting between two strings.")
+
+(defun mct-sort-by-alpha (completions)
+  "Sort COMPLETIONS alphabetically.
 This function can be used as the value of the user option
 `completions-sort'."
   (sort
-   elements
-   (lambda (c1 c2)
-     (or (string-version-lessp c1 c2)
-         (< (length c1) (length c2))))))
+   completions
+   (lambda (string1 string2)
+     (funcall mct-sort-alpha-function string1 string2))))
 
-(defun mct-sort-by-history (elements)
-  "Sort ELEMENTS by minibuffer history, else return them unsorted.
+(defun mct-sort-by-alpha-then-by-length (completions)
+  "Sort COMPLETIONS first alphabetically, then by length.
 This function can be used as the value of the user option
 `completions-sort'."
-  (if-let ((hist (and (not (eq minibuffer-history-variable t))
-                      (symbol-value minibuffer-history-variable))))
-      (minibuffer--sort-by-position hist elements)
-    elements))
+  (sort
+   completions
+   (lambda (string1 string2)
+     (or (funcall mct-sort-alpha-function string1 string2)
+         (< (length string1) (length string2))))))
 
-(defun mct-sort-multi-category (elements)
-  "Sort ELEMENTS per completion category.
+(defalias 'mct-sort-by-alpha-length 'mct-sort-by-alpha-then-by-length
+  "Alias for `mct-sort-by-alpha-then-by-length'.")
+
+;; Based on `minibuffer-sort-by-history' from Emacs 30.
+(defun mct-sort-by-history (completions)
+  "Sort COMPLETIONS by minibuffer history, else return them unsorted.
 This function can be used as the value of the user option
 `completions-sort'."
-  (pcase (mct--completion-category)
-    ('kill-ring elements) ; no sorting
-    ('file (mct-sort-by-alpha-length elements))
-    (_ (mct-sort-by-history elements))))
+  (let ((alphabetized (mct-sort-by-alpha completions)))
+    ;; Only use history when it's specific to these completions.
+    (if (eq minibuffer-history-variable
+            (default-value minibuffer-history-variable))
+        alphabetized
+      (minibuffer--sort-by-position
+       (minibuffer--sort-preprocess-history minibuffer-completion-base)
+       alphabetized))))
+
+(defun mct-sort-by-directory-then-by-file (completions)
+  "Sort COMPLETIONS with `mct-sort-by-alpha-then-by-length' with directories first.
+Also remove the implicit directories . and .. from the COMPLETIONS."
+  (setq completions (seq-remove (lambda (x) (or (string= "./" x) (string= "../" x))) completions))
+  (setq completions (mct-sort-by-alpha completions))
+  (nconc (seq-filter (lambda (x) (string-suffix-p "/" x)) completions)
+         (seq-remove (lambda (x) (string-suffix-p "/" x)) completions)))
+
+(defalias 'mct-sort-directories-then-files 'mct-sort-by-directory-then-by-file
+  "Alias for `mct-sort-by-directory-then-by-file'.")
+
+(defun mct--sort-multi-category-get-function (function-and-category)
+  "Return the first sort function among FUNCTION-AND-CATEGORY."
+  (catch 'found
+    (dolist (element function-and-category)
+      (pcase-dolist (`(,symbols . ,sort) mct-sort-by-command-or-category)
+        (when (if (listp symbols)
+                  (memq element symbols)
+                (eq element symbols))
+          (throw 'found sort))))))
+
+(defun mct-sort-multi-category (completions)
+  "Sort COMPLETIONS per command or completion category.
+Do it in accordance with the user option `mct-sort-by-command-or-category'."
+  (if-let* ((symbols (list (mct--this-command) (mct--completion-category)))
+            (sort-fn (or (mct--sort-multi-category-get-function symbols)
+                         (alist-get t mct-sort-by-command-or-category)))
+            (_ (functionp sort-fn)))
+      (funcall sort-fn completions)
+    completions))
 
 ;;;; Basics of intersection between minibuffer and Completions buffer
 
@@ -368,9 +422,39 @@ Can be used in `mct-completion-window-size'."
   "Fit Completions buffer to its window."
   (when-let* ((window (mct--get-completion-window))
               (size mct-completion-window-size)
-              (max (car size))
-              (min (cdr size)))
-    (fit-window-to-buffer window (mct--height max) (mct--height min))))
+              (max-height (mct--height (car size)))
+              (min-height (mct--height (cdr size)))
+              ;; For vertical alignment of contents and horizontal
+              ;; borders of completion window when content size do
+              ;; not reaches the upper bound of the window height.
+              (window-resize-pixelwise t)
+
+              ;; `fit-window-to-buffer' calculates the an upper
+              ;; bound of window height as the minimum of provided
+              ;; `max-height' and the sum of the content size and
+              ;; some pixels border lines, which results in the
+              ;; final window height lacking for some pixels when
+              ;; the content size larger than `max-height'.  So we
+              ;; need some tricks to align contents and window
+              ;; horizontal borders for this case.  The following
+              ;; few lines calculate the amount of lacked pixels,
+              ;; `delta-px', and enlarge window height if in the
+              ;; case.
+
+              (frame (window-frame window))
+              (line-height-px (window-default-line-height window))
+              ;; `max-height' in pixels
+              (max-height-px (* line-height-px (mct--height max-height)))
+              ;; Current window height, including all the stuffs.
+              (height-px (+ (cdr (window-text-pixel-size
+	                              window nil t nil (frame-pixel-height frame) t))
+                            (window-scroll-bar-height window)
+                            (window-bottom-divider-width window)))
+              (mode-line-height-px (window-mode-line-height window))
+              (delta-px (% mode-line-height-px line-height-px))) ; Offset
+    (fit-window-to-buffer window max-height min-height)
+    (when (< max-height-px height-px)
+      (window-resize window delta-px nil window window-resize-pixelwise))))
 
 (defun mct--minimum-input ()
   "Test for minimum requisite input for live completions.
@@ -457,7 +541,7 @@ Meant to be added to `after-change-functions'."
   "Return t if Mct is active."
   (when-let* ((win (active-minibuffer-window))
               (buf (window-buffer win)))
-      (buffer-local-value 'mct--active buf)))
+    (buffer-local-value 'mct--active buf)))
 
 (defun mct--minibuffer-completion-help-advice (&rest app)
   "Prepare APP advice around `display-completion-list'."
@@ -493,7 +577,7 @@ Apply APP by first setting up the minibuffer to work with Mct."
 (defun mct-focus-minibuffer ()
   "Focus the active minibuffer."
   (interactive nil mct-mode)
-  (when-let ((mini (active-minibuffer-window)))
+  (when-let* ((mini (active-minibuffer-window)))
     (select-window mini)))
 
 (defun mct--get-completion-window ()
@@ -522,7 +606,7 @@ command will first move per invocation to the former, then the
 latter, and then continue to switch between the two.
 
 The continuous switch is essentially the same as running
-`mct-focus-minibuffer' and `switch-to-Completions in
+`mct-focus-minibuffer' and `switch-to-completions' in
 succession.
 
 What constitutes a Completions window is ultimately determined
@@ -560,8 +644,7 @@ by `mct--completions-window-name'."
   "Return the `point' of the first completion."
   (save-excursion
     (goto-char (point-max))
-    (when completions-header-format
-     (next-completion 1))
+    (next-completion 1)
     (point)))
 
 (defun mct--last-completion-point ()
@@ -671,10 +754,7 @@ the minibuffer."
 If ARG is supplied, move that many completion groups at a time."
   (interactive "p" mct-mode)
   (dotimes (_ (or arg 1))
-    (when-let (group (save-excursion
-                       (text-property-search-forward 'face
-                                                     'completions-group-separator
-                                                     t nil)))
+    (when-let* ((group (save-excursion (text-property-search-forward 'face 'completions-group-separator t nil))))
       (let ((pos (prop-match-end group)))
         (unless (eq pos (point-max))
           (goto-char pos)
@@ -687,10 +767,7 @@ If ARG is supplied, move that many completion groups at a time."
   (dotimes (_ (or arg 1))
     ;; skip back, so if we're at the top of a group, we go to the previous one...
     (forward-line -1)
-    (if-let (group (save-excursion
-                     (text-property-search-backward 'face
-                                                    'completions-group-separator
-                                                    t nil)))
+    (if-let* ((group (save-excursion (text-property-search-backward 'face 'completions-group-separator t nil))))
         (let ((pos (prop-match-beginning group)))
           (unless (eq pos (point-min))
             (goto-char pos)
@@ -709,7 +786,7 @@ If ARG is supplied, move that many completion groups at a time."
   (interactive nil mct-mode)
   (choose-completion)
   (when (active-minibuffer-window)
-    (minibuffer-force-complete-and-exit)))
+    (exit-minibuffer)))
 
 (defun mct-choose-completion-no-exit ()
   "Run `choose-completion' in the Completions without exiting."
@@ -720,6 +797,21 @@ If ARG is supplied, move that many completion groups at a time."
 (defvar crm-completion-table)
 (defvar crm-separator)
 
+;; This is adapted from the README of the `vertico' package by Daniel
+;; Mendler.
+(defun mct--crm-indicator (args)
+  "Display an indicator for `completing-read-multiple' based on ARGS."
+  (cons (format "%s %s | %s"
+                (propertize "`completing-read-multiple' separator is" 'face 'shadow)
+                (propertize
+                 (format " %s "
+                         (replace-regexp-in-string
+                          "\\`\\[.*?]\\*\\|\\[.*?]\\*\\'" ""
+                          crm-separator))
+                 'face '(highlight bold))
+                (car args))
+        (cdr args)))
+
 (defun mct--regex-to-separator (regex)
   "Parse REGEX of `crm-separator' in `mct-choose-completion-dwim'."
   (save-match-data
@@ -769,8 +861,8 @@ determined as follows:
 A candidate is recognised for as long as point is not past its
 last character."
   (interactive nil mct-mode)
-  (when-let ((window (mct--get-completion-window))
-             ((active-minibuffer-window)))
+  (when-let* ((window (mct--get-completion-window))
+              ((active-minibuffer-window)))
     (with-current-buffer (window-buffer window)
       (let* ((old-point (save-excursion
                           (select-window window)
@@ -821,8 +913,7 @@ this command is then required to abort the session."
 ;;;;; Stylistic tweaks and refinements
 
 ;; Note that this solves bug#45686:
-;; <https://debbugs.gnu.org/cgi/bugreport.cgi?bug=45686>
-;; TODO review that stealthily does not affect the region mode, it seems intrusive.
+;; <https://debbugs.gnu.org/cgi/bugreport.cgi?bug=45686>.
 (defun mct--stealthily (&rest app)
   "Prevent minibuffer default from counting as a modification.
 Apply APP while inhibiting modification hooks."
@@ -836,6 +927,13 @@ Apply APP while inhibiting modification hooks."
   (when completions-header-format
     (setq-local display-line-numbers-offset -1)))
 
+(defun mct--completion--insert-strings (&rest _)
+  "Make `completion--insert-strings' not show the truncation message."
+  (when-let* ((_ (bound-and-true-p completions--lazy-insert-button))
+              (button completions--lazy-insert-button))
+    ;; (replace-region-contents (button-start button) (button-end button) "​"))) ; zero-width space here
+    (overlay-put (car (overlays-at (button-start button))) 'invisible t)))
+
 ;;;;; Shadowed path
 
 ;; Adapted from icomplete.el
@@ -866,19 +964,19 @@ Apply APP while inhibiting modification hooks."
   "Priority used on the `mct--highlight-overlay'.
 This value means that it is overriden by the active region.")
 
-;; The `if-let' is to prevent highlighting of empty space, such as by
+;; The `if-let*' is to prevent highlighting of empty space, such as by
 ;; clicking on it with the mouse.
 (defun mct--completions-completion-beg ()
   "Return point of completion candidate at START and END."
-  (if-let ((string (mct--completion-at-point-p)))
+  (if-let* ((string (mct--completion-at-point-p)))
       (save-excursion
         (prop-match-beginning (text-property-search-forward 'completion--string)))
     (point)))
 
-;; Same as above for the `if-let'.
+;; Same as above for the `if-let*'.
 (defun mct--completions-completion-end ()
   "Return end of completion candidate."
-  (if-let ((string (mct--completion-at-point-p)))
+  (if-let* ((string (mct--completion-at-point-p)))
       (save-excursion
         (1+ (line-end-position)))
     (point)))
@@ -913,8 +1011,10 @@ This value means that it is overriden by the active region.")
     (define-key map [remap keyboard-quit] #'mct-keyboard-quit-dwim)
     (define-key map [remap next-line] #'mct-next-completion-or-mini)
     (define-key map (kbd "n") #'mct-next-completion-or-mini)
+    (define-key map (kbd "<down>") #'mct-next-completion-or-mini)
     (define-key map [remap previous-line] #'mct-previous-completion-or-mini)
     (define-key map (kbd "p") #'mct-previous-completion-or-mini)
+    (define-key map (kbd "<up>") #'mct-previous-completion-or-mini)
     (define-key map [remap backward-paragraph] #'mct-previous-completion-group)
     (define-key map [remap forward-paragraph] #'mct-next-completion-group)
     (define-key map (kbd "M-p") #'mct-previous-completion-group)
@@ -933,11 +1033,14 @@ This value means that it is overriden by the active region.")
     (define-key map (kbd "C-j") #'exit-minibuffer)
     (define-key map [remap next-line] #'mct-switch-to-completions-top)
     (define-key map [remap next-line-or-history-element] #'mct-switch-to-completions-top)
+    (define-key map (kbd "<down>") #'mct-switch-to-completions-top)
     (define-key map [remap previous-line] #'mct-switch-to-completions-bottom)
-    (define-key map [remap previous-line-or-history-element] #'mct-switch-to-completions-bottom)
+    (define-key map (kbd "<up>") #'mct-switch-to-completions-bottom)
     (define-key map (kbd "M-e") #'mct-edit-completion)
     (define-key map (kbd "C-<return>") #'mct-complete-and-exit)
     (define-key map (kbd "C-l") #'mct-list-completions-toggle)
+    (define-key map (kbd "<left>") nil)
+    (define-key map (kbd "<right>") nil)
     map)
   "Derivative of `minibuffer-local-completion-map'.")
 
@@ -968,57 +1071,20 @@ This value means that it is overriden by the active region.")
 
 ;;;;; Dynamic completion
 
-(defun mct--persist-dynamic-completion (&rest _)
-  "Persist completion, per `mct-persist-dynamic-completion'."
-  (when (and (not (mct--symbol-in-list mct-completion-blocklist))
-             mct-persist-dynamic-completion
-             (memq (mct--completion-category) mct--dynamic-completion-categories)
-             mct-live-completion)
+(defun mct-persist-completions-buffer (&rest _)
+  "Persist the completions buffer if there are still candidates.
+Do this under any of the following conditions:
+
+- The command is in the `mct-completion-passlist'.
+
+- The `mct-live-completion' is non-nil."
+  (when (and (not (mct--blocklist-p))
+             (or (mct--passlist-p) mct-live-completion))
     (mct-focus-minibuffer)
     (mct--show-completions)))
 
-(defun mct--setup-dynamic-completion-persist ()
-  "Set up `mct-persist-dynamic-completion'."
-  (let ((commands '(choose-completion minibuffer-complete minibuffer-force-complete)))
-    (if (bound-and-true-p mct-mode)
-        (dolist (fn commands)
-          (advice-add fn :after #'mct--persist-dynamic-completion))
-      (dolist (fn commands)
-        (advice-remove fn #'mct--persist-dynamic-completion)))))
-
 ;;;;; mct-mode declaration
 
-(declare-function minibuf-eldef-setup-minibuffer "minibuf-eldef")
-
-;;;###autoload
-(define-minor-mode mct-mode
-  "Set up opinionated default completion UI."
-  :global t
-  :group 'mct
-  (if mct-mode
-      (progn
-        (add-hook 'completion-list-mode-hook #'mct--setup-completion-list)
-        (add-hook 'minibuffer-setup-hook #'mct--setup-passlist)
-        (advice-add #'completing-read-default :around #'mct--completing-read-advice)
-        (advice-add #'completing-read-multiple :around #'mct--completing-read-advice)
-        (advice-add #'minibuffer-completion-help :around #'mct--minibuffer-completion-help-advice)
-        (advice-add #'minibuf-eldef-setup-minibuffer :around #'mct--stealthily))
-    (remove-hook 'completion-list-mode-hook #'mct--setup-completion-list)
-    (remove-hook 'minibuffer-setup-hook #'mct--setup-passlist)
-    (advice-remove #'completing-read-default #'mct--completing-read-advice)
-    (advice-remove #'completing-read-multiple #'mct--completing-read-advice)
-    (advice-remove #'minibuffer-completion-help #'mct--minibuffer-completion-help-advice)
-    (advice-remove #'minibuf-eldef-setup-minibuffer #'mct--stealthily))
-  (mct--setup-dynamic-completion-persist)
-  (mct--setup-message-advices))
-
-(define-obsolete-function-alias
-  'mct-minibuffer-mode
-  'mct-mode
-  "1.0.0")
-
-(make-obsolete 'mct-region-mode nil "1.0.0")
-
 ;; Adapted from Omar Antolín Camarena's live-completions library:
 ;; <https://github.com/oantolin/live-completions>.
 (defun mct--honor-inhibit-message (&rest app)
@@ -1035,10 +1101,30 @@ This value means that it is overriden by the active region.")
         (apply app))
     (apply app)))
 
-(defun mct--setup-message-advices ()
-  "Silence the minibuffer and the Completions."
+(defvar mct-last-completions-sort-value (bound-and-true-p completions-sort)
+  "Last value of `completions-sort'.")
+
+(declare-function minibuf-eldef-setup-minibuffer "minibuf-eldef")
+
+;;;###autoload
+(define-minor-mode mct-mode
+  "Set up opinionated default completion UI."
+  :global t
+  :group 'mct
   (if mct-mode
       (progn
+        (setq completions-sort #'mct-sort-multi-category)
+        (add-hook 'completion-list-mode-hook #'mct--setup-completion-list)
+        (add-hook 'minibuffer-setup-hook #'mct--setup-passlist)
+        (advice-add #'completing-read-default :around #'mct--completing-read-advice)
+        (advice-add #'completing-read-multiple :around #'mct--completing-read-advice)
+        (when (and mct-completing-read-multiple-indicator (< emacs-major-version 31))
+          (advice-add #'completing-read-multiple :filter-args #'mct--crm-indicator))
+        (advice-add #'minibuffer-completion-help :around #'mct--minibuffer-completion-help-advice)
+        (advice-add #'minibuf-eldef-setup-minibuffer :around #'mct--stealthily)
+        (advice-add #'completion--insert-strings :after #'mct--completion--insert-strings)
+        (dolist (fn '(choose-completion minibuffer-complete minibuffer-force-complete))
+          (advice-add fn :after #'mct-persist-completions-buffer))
         (dolist (fn '(exit-minibuffer
                       choose-completion
                       minibuffer-force-complete
@@ -1046,6 +1132,17 @@ This value means that it is overriden by the active region.")
                       minibuffer-force-complete-and-exit))
           (advice-add fn :around #'mct--messageless))
         (advice-add #'minibuffer-message :around #'mct--honor-inhibit-message))
+    (setq completions-sort mct-last-completions-sort-value)
+    (remove-hook 'completion-list-mode-hook #'mct--setup-completion-list)
+    (remove-hook 'minibuffer-setup-hook #'mct--setup-passlist)
+    (advice-remove #'completing-read-default #'mct--completing-read-advice)
+    (advice-remove #'completing-read-multiple #'mct--completing-read-advice)
+    (advice-remove #'completing-read-multiple #'mct--crm-indicator)
+    (advice-remove #'minibuffer-completion-help #'mct--minibuffer-completion-help-advice)
+    (advice-remove #'minibuf-eldef-setup-minibuffer #'mct--stealthily)
+    (advice-remove #'completion--insert-strings #'mct--completion--insert-strings)
+    (dolist (fn '(choose-completion minibuffer-complete minibuffer-force-complete))
+      (advice-remove fn #'mct-persist-completions-buffer))
     (dolist (fn '(exit-minibuffer
                   choose-completion
                   minibuffer-force-complete
@@ -1054,5 +1151,12 @@ This value means that it is overriden by the active region.")
       (advice-remove fn #'mct--messageless))
     (advice-remove #'minibuffer-message #'mct--honor-inhibit-message)))
 
+(define-obsolete-function-alias
+  'mct-minibuffer-mode
+  'mct-mode
+  "1.0.0")
+
+(make-obsolete 'mct-region-mode nil "1.0.0")
+
 (provide 'mct)
 ;;; mct.el ends here
